#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from io import BytesIO
import json
import os
from typing import Any, Dict, List, Optional
from zipfile import ZipFile

import trafaret as t

from datarobot.enums import ListCustomTemplatesSortQueryParams
from datarobot.models.api_object import APIObject
from datarobot.utils import from_api, to_api
from datarobot.utils.pagination import unpaginate


def file_content(file: str) -> bytes:
    """
    Converts a file/directory name (or byte string) to a byte string.

    When 'file' is a directory, the content is a ZIP of the directory.
    When 'file' is a file, the content is the file contents and should be a ZIP file (but not checked).
    Otherwise, the content is the encoded [assumed] string.
    """
    if os.path.isdir(file):
        temp_file = BytesIO()
        with ZipFile(temp_file, "w") as zipf:
            for dir_path, _, files in os.walk(file):
                for f in files:
                    fn = os.path.join(dir_path, f)
                    archive_path = os.path.relpath(fn, file)
                    zipf.write(fn, arcname=archive_path)
        content = temp_file.getvalue()
    elif os.path.isfile(file):
        with open(file, "rb") as fp:
            content = fp.read()
    else:
        content = bytes(file.encode("utf-8"))

    return content


def to_json_str(obj: Any) -> str:
    """
    Converts a Python object to a JSON string.

    This avoids redundant logic to convert APIObject|str|Any to a json encoded string.
    """
    if isinstance(obj, APIObject):
        api_obj = to_api(obj)
    elif isinstance(obj, str):
        api_obj = json.loads(obj)
    else:
        api_obj = obj
    return json.dumps(api_obj)


class DefaultEnvironment(APIObject):
    """
    Default execution environment.
    """

    _converter = t.Dict(
        {
            t.Key("environment_id"): t.String(),
            t.Key("environment_version_id"): t.String(),
        }
    ).ignore_extra("*")
    schema = _converter

    def __init__(self, environment_id: str, environment_version_id: str):
        self.environment_id = environment_id
        self.environment_version_id = environment_version_id

    def __repr__(self) -> str:
        return "{}(environment_id={}, environment_version_id={})".format(
            self.__class__.__name__, self.environment_id, self.environment_version_id
        )


class CustomMetricMetadata(APIObject):
    """
    Metadata for custom metrics.
    """

    _converter = t.Dict(
        {
            # NOTE: several values are really enums, but treated as strings for simplicity
            t.Key("units"): t.String(),
            t.Key("directionality"): t.String(),
            t.Key("type"): t.String(),
            t.Key("time_step"): t.String(),
            t.Key("is_model_specific"): t.Bool(),
            t.Key("template_metric_type", optional=True): t.Or(t.Null(), t.String()),
        }
    ).ignore_extra("*")
    schema = _converter

    def __init__(
        self,
        units: str,
        directionality: str,
        type: str,
        time_step: str,
        is_model_specific: bool,
        template_metric_type: Optional[str] = None,
    ):
        self.units = units
        self.directionality = directionality
        self.type = type
        self.time_step = time_step
        self.is_model_specific = is_model_specific
        self.template_metric_type = template_metric_type


class TemplateMetadata(APIObject):
    """
    Metadata for the custom templates.
    """

    _converter = t.Dict(
        {
            t.Key("readme", optional=True): t.Or(t.Null(), t.String()),
            t.Key("source", optional=True): t.Or(t.Null(), t.Dict().allow_extra("*")),
            t.Key("tags", optional=True): t.List(t.String()),
            t.Key("custom_metric_metadata", optional=True): t.Or(
                t.Null(), CustomMetricMetadata.schema
            ),
            t.Key("feature_flag", optional=True): t.Or(t.Null(), t.String()),
            t.Key("preview_image", optional=True): t.Or(t.Null(), t.String()),
        }
    ).ignore_extra("*")
    schema = _converter

    def __init__(
        self,
        readme: Optional[str] = None,
        source: Optional[str] = None,
        tags: Optional[List[str]] = None,
        feature_flag: Optional[str] = None,
        preview_image: Optional[str] = None,
        custom_metric_metadata: Optional[CustomMetricMetadata] = None,
    ):
        self.readme = readme
        self.source = source
        self.tags = tags
        self.feature_flag = feature_flag
        self.preview_image = preview_image
        self.custom_metric_metadata = custom_metric_metadata


class CustomTemplate(APIObject):
    """
    Template for custom activity (e.g. custom-metrics, applications).
    """

    _path = "customTemplates/"

    _converter = t.Dict(
        {
            t.Key("default_environment"): DefaultEnvironment.schema,
            t.Key("default_resource_bundle_id", optional=True, default=None): t.Or(
                t.Null(), t.String()
            ),
            t.Key("description"): t.String(),
            t.Key("enabled"): t.Bool(),
            t.Key("id"): t.String(),
            t.Key("items"): t.List(t.Dict().allow_extra("*")),
            t.Key("name"): t.String(),
            t.Key("template_metadata"): TemplateMetadata.schema,
            t.Key("template_sub_type"): t.String(),
            t.Key("template_type"): t.String(),
            t.Key("is_hidden", optional=True, default=None): t.Or(t.Bool(), t.Null()),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id: str,
        name: str,
        description: str,
        template_metadata: Dict[str, Any],
        default_environment: Dict[str, Any],
        default_resource_bundle_id: Optional[str],
        items: List[Dict[str, Any]],
        template_type: str,
        template_sub_type: str,
        enabled: bool,
        is_hidden: bool,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.template_metadata = template_metadata
        self.default_environment = default_environment
        self.default_resource_bundle_id = default_resource_bundle_id
        self.items = items
        self.template_type = template_type
        self.template_sub_type = template_sub_type
        self.enabled = enabled
        self.is_hidden = is_hidden

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id!r}, name={self.name!r})"

    @classmethod
    def create(
        cls,
        name: str,
        description: str,
        template_type: str,
        template_sub_type: str,
        template_metadata: TemplateMetadata | dict[str, Any] | str,
        default_environment: DefaultEnvironment | dict[str, Any] | str,
        file: str,
        default_resource_bundle_id: Optional[str] = None,
        enabled: Optional[bool] = None,
        is_hidden: Optional[bool] = None,
    ) -> CustomTemplate:
        """
        Create the custom template.

        .. versionadded:: v3.9

        Parameters
        ----------
        name: str
            The template name.
        description: str
            A description of the template.
        template_type: str
            The template type.
        template_sub_type: str
            The template sub-type.
        template_metadata: TemplateMetadata|dict[str, Any]|str
            The metadata associated with the template, provided as TemplateMetadata or a JSON-encoded string.
        default_environment: DefaultEnvironment|dict[str, Any]|str
            The default environment associated with the template, provided as DefaultEnvironment or a JSON-encoded
            string.
        file: str
            The path to the template directory or file, or the contents of the template.
        default_resource_bundle_id: Optional[str]
            The default resource bundle ID.
        enabled: Optional[bool]
            Whether the template is enabled (default is true).
        is_hidden: Optional[bool]
            Whether the template is hidden (default is false).

        Examples
        --------
        .. code-block:: python

            from datarobot import CustomTemplate
            from datarobot.models.custom_templates import DefaultEnvironment
            def_env = DefaultEnvironment(
                environment_id='679d47c8ce1ecd17326f3fdf',
                environment_version_id='679d47c8ce1ecd17326f3fe3',
            )
            template = template.create(
                name="My new template",
                default_environment=def_env,
                description='Updated template with environment v17',
            )
        """

        body = {}
        body["name"] = name
        body["description"] = description
        if default_resource_bundle_id:
            body["defaultResourceBundleId"] = default_resource_bundle_id
        body["templateType"] = template_type
        body["templateSubType"] = template_sub_type
        body["templateMetadata"] = to_json_str(template_metadata)
        body["defaultEnvironment"] = to_json_str(default_environment)
        if enabled is not None:
            body["enabled"] = str(enabled).lower()
        if is_hidden is not None:
            body["isHidden"] = str(is_hidden).lower()

        url = f"{cls._path}/"
        response = cls._client.build_request_with_file(
            method="POST",
            url=url,
            form_data=body,
            fname="template-content",
            content=file_content(file),
        )

        # the create response only has the customTemplateId, so use that to fetch the object
        template_id = response.json().get("customTemplateId")
        return CustomTemplate.get(template_id)

    @classmethod
    def list(
        cls,
        search: Optional[str] = None,
        order_by: Optional[ListCustomTemplatesSortQueryParams] = None,
        tag: Optional[str] = None,
        template_type: Optional[str] = None,
        template_sub_type: Optional[str] = None,
        publisher: Optional[str] = None,
        category: Optional[str] = None,
        show_hidden: Optional[bool] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List[CustomTemplate]:
        """List all custom templates.

        .. versionadded:: v3.7

        Parameters
        ----------
        search: Optional[str]
            Search string.
        order_by: Optional[ListCustomTemplatesSortQueryParams]
            Ordering field.
        tag: Optional[str]
            Tag associated with the template.
        template_type: Optional[str]
            Type of the template.
        template_type: Optional[str]
            Sub-type of the template.
        publisher: Optional[str]
            Only return custom templates with this publisher.
        category: Optional[str]
            Only return custom templates with this category (use case).
        show_hidden: Optional[bool]
            Whether the template is hidden (default is false).
        offset: Optional[int]
            Offset for pagination.
        limit: Optional[int]
            Limit for pagination.

        Returns
        -------
        templates: List[CustomTemplate]
        """
        params: Dict[str, Any] = {}
        if search:
            params["search"] = search
        if order_by:
            params["orderBy"] = order_by
        if tag:
            params["tag"] = tag
        if template_type:
            params["templateType"] = template_type
        if template_sub_type:
            params["templateSubType"] = template_sub_type
        if publisher:
            params["publisher"] = publisher
        if category:
            params["category"] = category
        if show_hidden is not None:
            params["showHidden"] = show_hidden
        if offset:
            params["offset"] = offset
        if limit:
            params["limit"] = limit

        if offset is None:
            data = unpaginate(cls._path, params, cls._client)
        else:
            data = cls._client.get(cls._path, params=params if params else None).json()["data"]
        return [cls.from_server_data(d) for d in data]

    @classmethod
    def get(cls, template_id: str) -> CustomTemplate:
        """Get a custom template by ID.

        .. versionadded:: v3.7

        Parameters
        ----------
        template_id: str
            ID of the template.

        Returns
        -------
        template : CustomTemplate
        """
        response = cls._client.get(f"{cls._path}{template_id}/")
        return cls.from_server_data(response.json())

    def update(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        default_resource_bundle_id: Optional[str] = None,
        template_type: Optional[str] = None,
        template_sub_type: Optional[str] = None,
        template_metadata: Optional[TemplateMetadata | str] = None,
        default_environment: Optional[DefaultEnvironment | str] = None,
        file: Optional[str] = None,
        enabled: Optional[bool] = None,
        is_hidden: Optional[bool] = None,
    ) -> None:
        """
        Update the custom template.

        .. versionadded:: v3.7

        Parameters
        ----------
        name: Optional[str]
            The template name.
        description: Optional[str]
            A description of the template.
        default_resource_bundle_id: Optional[str]
            The default resource bundle ID.
        template_type: Optional[str]
            The template type.
        template_sub_type: Optional[str]
            The template sub-type.
        template_metadata: Optional[TemplateMetadata|str]
            The metadata associated with the template, provided as TemplateMetadata or a JSON encoded string.
        default_environment: Optional[DefaultEnvironment|str]
            The default environment associated with the template, provided as DefaultEnvironment or a JSON encoded
            string.
        file: str
            The path to the template directory or file, or the contents of the template.
        enabled: Optional[bool]
            Whether the template is enabled (default is true).
        is_hidden: Optional[bool]
            Whether the template is hidden (default is false).

        Examples
        --------
        .. code-block:: python

            from datarobot import CustomTemplate
            from datarobot.models.custom_templates import DefaultEnvironment
            new_env = DefaultEnvironment(
                environment_id='679d47c8ce1ecd17326f3fdf',
                environment_version_id='679d47c8ce1ecd17326f3fe3',
            )
            template = CustomTemplate.get(template_id='5c939e08962d741e34f609f0')
            template.update(default_environment=new_env, description='Updated template with environment v17')
        """

        _default_env_obj: Any = None
        _template_meta_obj: Any = None
        body = {}
        if name:
            body["name"] = name
        if description:
            body["description"] = description
        if default_resource_bundle_id:
            body["defaultResourceBundleId"] = default_resource_bundle_id
        if template_type:
            body["templateType"] = template_type
        if template_sub_type:
            body["templateSubType"] = template_sub_type
        if template_metadata:
            if isinstance(template_metadata, TemplateMetadata):
                api_obj = to_api(template_metadata)
                _template_meta_obj = from_api(api_obj)  # type: ignore[arg-type]
                template_metadata = json.dumps(api_obj)
            else:
                _template_meta_obj = from_api(json.loads(template_metadata))
            body["templateMetadata"] = template_metadata
        if default_environment:
            if isinstance(default_environment, DefaultEnvironment):
                api_obj = to_api(default_environment)
                _default_env_obj = from_api(api_obj)  # type: ignore[arg-type]
                default_environment = json.dumps(api_obj)
            else:
                _default_env_obj = from_api(json.loads(default_environment))
            body["defaultEnvironment"] = default_environment
        if enabled is not None:
            body["enabled"] = str(enabled).lower()
        if is_hidden is not None:
            body["isHidden"] = str(is_hidden).lower()

        if not body and not file:
            raise ValueError("Nothing to update")

        url = f"{self._path}{self.id}/"
        items = None
        if not file:
            self._client.patch(url, data=body)
        else:
            response = self._client.build_request_with_file(
                method="PATCH",
                url=url,
                form_data=body,
                fname="template-content",
                content=file_content(file),
            )
            updated = self.from_server_data(response.json())
            items = updated.items

        # update the local data
        if name:
            self.name = name
        if description:
            self.description = description
        if default_resource_bundle_id:
            self.default_resource_bundle_id = default_resource_bundle_id
        if template_type:
            self.template_type = template_type
        if template_sub_type:
            self.template_sub_type = template_sub_type
        if _template_meta_obj:
            self.template_metadata = _template_meta_obj
        if _default_env_obj:
            self.default_environment = _default_env_obj
        if enabled is not None:
            self.enabled = enabled
        if is_hidden is not None:
            self.is_hidden = is_hidden
        if items is not None:
            self.items = items

    def delete(self) -> None:
        """
        Delete this custom template.

        .. versionadded:: v3.7

        Returns
        -------
        None
        """

        url = f"{self._path}{self.id}/"
        self._client.delete(url)

    def download_content(
        self, index: Optional[int] = None, filename: Optional[str] = None
    ) -> Optional[bytes]:
        """
        Retrieve the file content for the given item.

        The item can be identified by filename or index in the array of items.

        .. versionadded:: v3.9

        Parameters
        ----------
        filename: Optional[str]
            The file name to retrieve.
        index: Optional[int]
            Index of the item to retrieve.

        Returns
        -------
        Bytes content of the file.
        """
        item = None
        if index is not None:
            if index > len(self.items):
                raise ValueError(
                    f"Index out of range -- only {len(self.items)} items are available"
                )
            item = self.items[index]
        elif filename is not None:
            item = next((i for i in self.items if i.get("name") == filename), None)
            if not item:
                names = [str(_item.get("name")) for _item in self.items]
                raise ValueError(f"File {filename} not found in: {', '.join(names)}")
        elif len(self.items) == 1:
            item = self.items[0]
        else:
            raise ValueError("Item must be specified by index or filename")

        url = f"{self._path}{self.id}/files/{item.get('id')}/"
        resp = self._client.get(url)
        return str(resp.json().get("content")).encode("utf-8") if resp.ok else None



#
# Copyright 2024-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import trafaret as t

from datarobot.enums import AggregationType, enum_to_list
from datarobot.models.api_object import APIObject
from datarobot.models.genai.insights_configuration import InsightsConfiguration
from datarobot.utils.pagination import unpaginate

aggregation_value_trafaret = t.Dict(
    {
        t.Key("item"): t.String,
        t.Key("value"): t.Float,
    }
).ignore_extra("*")

evaluation_dataset_metric_aggregation_trafaret = t.Dict(
    {
        t.Key("llm_blueprint_id"): t.String,
        t.Key("evaluation_dataset_configuration_id", optional=True, default=None): t.Or(
            t.String, t.Null
        ),
        t.Key("ootb_dataset_name", optional=True, default=None): t.Or(t.String, t.Null),
        t.Key("metric_name"): t.String(allow_blank=True),
        t.Key("dataset_id", optional=True, default=None): t.Or(t.String, t.Null),
        t.Key("dataset_name", optional=True, default=None): t.Or(
            t.String(allow_blank=True), t.Null
        ),
        t.Key("chat_id"): t.String,
        t.Key("chat_name"): t.String(allow_blank=True),
        t.Key(
            "aggregation_value",
            optional=True,
            default=None,
        ): t.Or(t.Float, t.List(aggregation_value_trafaret), t.Null),
        t.Key("aggregation_type"): t.Enum(*enum_to_list(AggregationType)),
        t.Key("creation_date"): t.String,
        t.Key("creation_user_id"): t.String,
        t.Key("tenant_id"): t.String,
        t.Key("custom_model_guard_id", optional=True, default=None): t.Or(t.String, t.Null),
    }
).ignore_extra("*")


class EvaluationDatasetMetricAggregation(APIObject):
    """Information about the aggregated metric results for one metric and one evaluation dataset.
    This class can list already computed aggregations or start the job computing the aggregations.
    Jobs will prompt an LLM blueprint or agentic workflow, compute metrics and aggregate the
    results across prompts.

    Attributes
    ----------
    llm_blueprint_id : str
        The LLM blueprint ID.
    evaluation_dataset_configuration_id : str | None
        The evaluation dataset configuration ID.
    ootb_dataset_name : str | None
        The name of the Datarobot-provided dataset that does not require additional configuration.
    metric_name : str
        The name of the metric.
    dataset_id : str | None
        The ID of the dataset used in the evaluation.
    dataset_name : str | None
        The name of the dataset used in the evaluation.
    chat_id : str
        The ID of the chat created to run the evaluation.
    chat_name : str
        The name of the chat that was created to run the evaluation.
    aggregation_value : float | List[Dict[str, float]] | None
        The aggregated metric result.
    aggregation_type : AggregationType
        The type of aggregation used for the metric results.
    creation_date : str
        The date the evaluation job was created.
    creation_user_id : str
        The ID of the user who created the evaluation job.
    tenant_id : str
        The ID of the tenant that owns the evaluation job.
    custom_model_guard_id : str | None
        The ID of the custom model's guard associated with the metric aggregation, if applicable.
    """

    _path = "api/v2/genai/evaluationDatasetMetricAggregations"
    _converter = evaluation_dataset_metric_aggregation_trafaret

    def __init__(
        self,
        llm_blueprint_id: str,
        evaluation_dataset_configuration_id: Optional[str],
        ootb_dataset_name: Optional[str],
        metric_name: str,
        dataset_id: Optional[str],
        dataset_name: Optional[str],
        chat_id: str,
        chat_name: str,
        aggregation_value: Optional[Union[float, List[Dict[str, float]]]],
        aggregation_type: AggregationType,
        creation_date: str,
        creation_user_id: str,
        tenant_id: str,
        custom_model_guard_id: Optional[str] = None,
    ) -> None:
        self.llm_blueprint_id = llm_blueprint_id
        self.evaluation_dataset_configuration_id = evaluation_dataset_configuration_id
        self.ootb_dataset_name = ootb_dataset_name
        self.metric_name = metric_name
        self.dataset_id = dataset_id
        self.dataset_name = dataset_name
        self.chat_id = chat_id
        self.chat_name = chat_name
        self.aggregation_value = aggregation_value
        self.aggregation_type = aggregation_type
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id
        self.tenant_id = tenant_id
        self.custom_model_guard_id = custom_model_guard_id

    @classmethod
    def create(
        cls,
        chat_name: str,
        llm_blueprint_ids: List[str],
        evaluation_dataset_configuration_id: str,
        insights_configuration: list[InsightsConfiguration],
    ) -> Any:
        """
        Create a new evaluation dataset metric aggregation job.  The job will run the
        specified metric for the specified LLM blueprint IDs using the prompt-response pairs in
        the evaluation dataset.

        Parameters
        ----------
        chat_name : str
            The name of the chat that will be created to run the evaluation in.
        llm_blueprint_ids : List[str]
            The LLM blueprint IDs to evaluate.
        evaluation_dataset_configuration_id : str
            The ID evaluation dataset configuration to use during the evaluation.
        insights_configuration : List[InsightsConfiguration]
            The insights configurations to use during the evaluation.

        Returns
        -------
        str
            The ID of the evaluation dataset metric aggregation job.
        """
        payload = {
            "chat_name": chat_name,
            "llm_blueprint_ids": llm_blueprint_ids,
            "evaluation_dataset_configuration_id": evaluation_dataset_configuration_id,
            "insights_configuration": [insight.to_dict() for insight in insights_configuration],
        }
        url = f"{cls._client.domain}/{cls._path}/"
        response_data = cls._client.post(url, json=payload)
        return response_data.json()["jobId"]

    @classmethod
    def list(
        cls,
        llm_blueprint_ids: Optional[List[str]] = None,
        chat_ids: Optional[List[str]] = None,
        evaluation_dataset_configuration_ids: Optional[List[str]] = None,
        metric_names: Optional[List[str]] = None,
        aggregation_types: Optional[List[str]] = None,
        current_configuration_only: Optional[bool] = False,
        sort: Optional[str] = None,
        offset: Optional[int] = 0,
        limit: Optional[int] = 100,
        non_errored_only: Optional[bool] = True,
    ) -> List[EvaluationDatasetMetricAggregation]:
        """List evaluation dataset metric aggregations.  The results will be filtered by the provided
        LLM blueprint IDs and chat IDs.

        Parameters
        ----------
        llm_blueprint_ids : List[str]
            The LLM blueprint IDs to filter on.
        chat_ids : List[str]
            The chat IDs to filter on.
        evaluation_dataset_configuration_ids : List[str]
            The evaluation dataset configuration IDs to filter on.
        metric_names : List[str]
            The metric names to filter on.
        aggregation_types : List[str]
            The aggregation types to filter on.
        current_configuration_only : Optional[bool]
            If True, only results that are associated with the current configuration of the LLM blueprint
            will be returned.  Defaults to False.
        sort : Optional[str]
            The field to sort on.  Defaults to None.
        offset : Optional[int]
            The offset to start at.  Defaults to 0.
        limit : Optional[int]
            The maximum number of results to return.  Defaults to 100.
        non_errored_only : Optional[bool]
            If True, only results that did not encounter an error will be returned.  Defaults to True.

        Returns
        -------
        List[EvaluationDatasetMetricAggregation]
            A list of evaluation dataset metric aggregations.
        """
        params = {
            "llm_blueprint_ids": llm_blueprint_ids,
            "chat_ids": chat_ids,
            "evaluation_dataset_configuration_ids": evaluation_dataset_configuration_ids,
            "metric_names": metric_names,
            "aggregation_types": aggregation_types,
            "current_configuration_only": current_configuration_only,
            "sort": sort,
            "offset": offset,
            "limit": limit,
            "non_errored_only": non_errored_only,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    @classmethod
    def delete(
        cls, llm_blueprint_ids: Optional[List[str]] = None, chat_ids: Optional[List[str]] = None
    ) -> None:
        """Delete the associated evaluation dataset metric aggregations.  Either llm_blueprint_ids
        or chat_ids must be provided.  If both are provided, only results matching both will be removed.

        Parameters
        ----------
        llm_blueprint_ids : List[str]
            The LLM blueprint IDs to filter on.
        chat_ids : List[str]
            The chat IDs to filter on.

        """
        if llm_blueprint_ids is None and chat_ids is None:
            raise ValueError("Either llm_blueprint_ids or chat_ids must be provided.")
        url = f"{cls._client.domain}/{cls._path}/"
        # parameters are camelcase here because delete doesn't seem to send them correctly otherwise.
        params = {
            "llmBlueprintIds": llm_blueprint_ids,
            "chatIds": chat_ids,
        }
        cls._client.delete(url, params=params)



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Optional

import trafaret as t

from datarobot._compat import TypedDict
from datarobot.models.api_object import APIObject

# TODO: We aren't handling `created` and `updated` for now
notebook_revision_trafaret = t.Dict(
    {
        t.Key("revision_id"): t.String,
        t.Key("notebook_id"): t.String,
        t.Key("name"): t.String,
        t.Key("is_auto"): t.Bool,
    }
).ignore_extra("*")


class CreateRevisionPayload(TypedDict, total=False):
    """
    Payload for creating a notebook revision.
    """

    name: Optional[str]
    notebook_path: Optional[str]
    is_auto: bool


class NotebookRevision(APIObject):
    """
    Represents a notebook revision.

    Attributes
    ----------

    revision_id : str
        The ID of the notebook revision.
    notebook_id : str
        The ID of the notebook.
    is_auto : bool
        Whether the revision was auto-saved.
    """

    _path = "notebookRevisions/"

    _converter = notebook_revision_trafaret

    def __init__(
        self,
        revision_id: str,
        notebook_id: str,
        name: str,
        is_auto: bool,
    ):
        self.revision_id = revision_id
        self.notebook_id = notebook_id
        self.name = name
        self.is_auto = is_auto

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(name="{self.name}", id={self.revision_id}, notebook_id={self.notebook_id})'

    @classmethod
    def create(
        cls, notebook_id: str, payload: Optional[CreateRevisionPayload] = None
    ) -> "NotebookRevision":
        """
        Create a new notebook revision.

        Parameters
        ----------
        notebook_id : str
            The ID of the notebook.
        payload : CreateRevisionPayload
            The payload to create the revision.

        Returns
        -------
        NotebookRevision
            Information about the created notebook revision.
        """
        r_data = cls._client.post(f"{cls._path}{notebook_id}/", data=payload or {})
        return cls.from_server_data(r_data.json())



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Dict, Optional

import trafaret as t

from datarobot.models.api_object import APIObject

notebook_user_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("activated"): t.Bool,
        t.Key("username"): t.String,
        t.Key("first_name"): t.String,
        t.Key("last_name"): t.String,
        t.Key("gravatar_hash", optional=True): t.String,
    }
).ignore_extra("*")


notebook_activity_trafaret = t.Dict(
    {
        t.Key("at"): t.String,
        t.Key("by"): notebook_user_trafaret,
    }
)


class NotebookUser(APIObject):
    """
    A user associated with a Notebook.

    Attributes
    ----------

    id : str
        The ID of the user.
    activated: bool
        Whether or not the user is enabled.
    username : str
        The username of the user, usually their email address.
    first_name : str
        The first name of the user.
    last_name : str
        The last name of the user.
    gravatar_hash : Optional[str]
        The gravatar hash of the user. Optional.
    tenant_phase : Optional[str]
        The phase that the user's tenant is in. Optional.
    """

    _converter = notebook_user_trafaret

    def __init__(
        self,
        id: str,
        activated: bool,
        username: str,
        first_name: str,
        last_name: str,
        gravatar_hash: Optional[str] = None,
        tenant_phase: Optional[str] = None,
    ):
        self.id = id
        self.activated = activated
        self.username = username
        self.first_name = first_name
        self.last_name = last_name
        self.gravatar_hash = gravatar_hash
        self.tenant_phase = tenant_phase


class NotebookActivity(APIObject):
    """
    A record of activity (i.e. last run, updated, etc.) in a Notebook.

    Attributes
    ----------

    at : str
        The time of the activity in the notebook.
    by : NotebookUser
        The user who performed the activity.
    """

    _converter = notebook_activity_trafaret

    def __init__(self, at: str, by: Dict[str, str]):
        self.at = at
        self.by = NotebookUser.from_server_data(by)



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations


class OAuthValidationErr(ValueError):
    """
    The base exception for all OAuth validation errors
    """


class OAuthProviderNotFound(OAuthValidationErr):
    pass


class OAuthFlowSessionExpired(OAuthValidationErr):
    """
    The OAuth flow session has expired
    """


class OAuthError(Exception):
    """
    The base exception for all OAuth flow errors
    TODO: Let's see if we have more specific errors derived from this one if not we should remove this class
    """


class OAuthFlowError(OAuthError):
    """
    The OAuth flow has failed
    """

    def __init__(self, provider_id: str, error_code: str, message: str) -> None:
        super().__init__(message)

        self.provider_id = provider_id
        self.error_code = error_code



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
# limitations under the License.
from __future__ import annotations

from datetime import datetime
import os
import platform
from types import TracebackType
from typing import TYPE_CHECKING, Any, Mapping, Sequence, Union
from urllib.parse import urljoin

import httpx
from pydantic import BaseModel, ConfigDict, Field

from datarobot._version import __version__
from datarobot.auth.datarobot.exceptions import (
    OAuthAuthorizationExpired,
    OAuthAuthorizationNotFound,
    OAuthServiceProviderNotFound,
    raise_http_exception,
)
from datarobot.auth.exceptions import OAuthFlowError, OAuthValidationErr
from datarobot.auth.oauth import (
    AsyncOAuthComponent,
    OAuthData,
    OAuthFlowSession,
    OAuthProvider,
    OAuthToken,
    Profile,
    SyncOAuthComponent,
)
from datarobot.auth.typing import Metadata
from datarobot.auth.utils import syncify
from datarobot.utils import camelize

if TYPE_CHECKING:
    import ssl


class _DRSchema(BaseModel):
    """
    The base schema that configures Pydantic to properly handle DataRobot API responses.
    """

    model_config = ConfigDict(
        alias_generator=camelize,
        populate_by_name=True,
        use_enum_values=True,
        extra="ignore",
    )


class _DROAuthProvider(_DRSchema):
    """
    DataRobot OAuth2 provider response schema.
    """

    id: str
    created_at: datetime
    updated_at: datetime
    name: str
    org_id: str | None = None
    type: str
    status: str
    client_id: str
    secure_config_id: str | None = None
    metadata: dict[str, Any] | None = Field(default_factory=lambda: {})

    def to_data(self) -> OAuthProvider:
        """
        Convert the DataRobot OAuth provider schema to the internal OAuthProvider model.
        """
        metadata = self.metadata or {}
        metadata.update(  # pylint: disable=no-member
            {
                "created_at": self.created_at,
                "updated_at": self.updated_at,
                "org_id": self.org_id,
                "secure_config_id": self.secure_config_id,
            }
        )

        return OAuthProvider(
            id=self.id,
            name=self.name,
            status=self.status,
            type=self.type,
            client_id=self.client_id,
            metadata=metadata,
        )


class _DROAuthRedirectData(_DRSchema):
    state: str
    redirect_url: str


class _DROAuthProviders(_DRSchema):
    data: list[_DROAuthProvider] = Field(default_factory=list)


class _DROAuthAuthorization(_DRSchema):
    id: str
    created_at: datetime
    user_id: str
    org_id: str | None = None
    status: str
    oauth_client_id: str
    credential_id: str
    refresh_token_expires_at: datetime | None = None
    provider: _DROAuthProvider


class _DROAuthTokenData(_DRSchema):
    """
    DataRobot OAuth token response schema.
    """

    access_token: str
    expires_at: datetime | None = None
    authorized_oauth_provider: _DROAuthAuthorization

    def to_data(self) -> OAuthToken:
        """
        Convert the DataRobot OAuth token schema to the internal OAuthToken model.
        """
        return OAuthToken(
            access_token=self.access_token,
            expires_at=self.expires_at,
        )


class _DROAuthUserProfile(_DRSchema):
    """
    DataRobot user profile response schema.
    """

    sub: str
    email: str
    name: str | None = None
    given_name: str | None = None
    family_name: str | None = None
    nickname: str | None = None
    description: str | None = None
    picture: str | None = None
    locale: str | None = None
    raw: Metadata | None = Field(default_factory=lambda: {})

    def to_data(self) -> Profile:
        """
        Convert the DataRobot user profile schema to the internal Profile model.
        """
        raw_data = {}

        if self.raw:
            raw_data = dict(self.raw)

        raw_data.update(
            {
                "description": self.description,
            }
        )

        return Profile(
            id=self.sub,
            email=self.email,
            name=self.name,
            first_name=self.given_name,
            last_name=self.family_name,
            preferred_username=self.nickname,
            picture=self.picture,
            locale=self.locale,
            metadata=raw_data,
        )


class _DROAuthData(_DRSchema):
    """
    DataRobot OAuth2 callback response schema.
    """

    authorized_provider: _DROAuthAuthorization
    user_info: _DROAuthUserProfile


class AsyncOAuth(AsyncOAuthComponent):
    """
    Asyncio OAuth2 implementation based on the DataRobot API.
    """

    def __init__(
        self,
        oauth_provider_ids: Sequence[str] = (),
        *,
        datarobot_endpoint: str | None = None,
        datarobot_api_token: str | None = None,
        http_client: httpx.AsyncClient | None = None,
        user_agent: str | None = None,
        follow_redirects: bool = True,
        timeout: float | httpx.Timeout | None = 5.0,
        ssl_verify: Union[
            bool, "ssl.SSLContext"
        ] = True,  # just "|" thrown "unsupported operand type" error
    ) -> None:
        """
        Initialize the AsyncOAuth client.

        Args:
            oauth_provider_ids: A sequence of OAuth provider IDs to register.
            datarobot_endpoint: The DataRobot API endpoint URL. If not provided,
                it will be read from the environment variable DATAROBOT_ENDPOINT.
            datarobot_api_token: The DataRobot API token. If not provided,
                it will be read from the environment variable DATAROBOT_API_TOKEN.
            http_client: An optional HTTP client to use for requests.
                If not provided, a new AsyncClient will be created.
            user_agent: An optional user agent string to use for requests.
            follow_redirects: Whether to follow redirects in HTTP requests.
            timeout: The timeout for HTTP requests.
            ssl_verify: Whether to verify SSL certificates or a custom SSL context.

        Raises:
            ValueError: If the DataRobot endpoint or API token is not provided.
        """
        if datarobot_endpoint is None:
            datarobot_endpoint = os.environ.get("DATAROBOT_ENDPOINT")

        if not datarobot_endpoint:
            raise ValueError(
                "DATAROBOT_ENDPOINT must be set in the environment or passed as an argument."
            )

        if datarobot_api_token is None:
            datarobot_api_token = os.environ.get("DATAROBOT_API_TOKEN")

        if not datarobot_api_token:
            raise ValueError(
                "DATAROBOT_API_TOKEN must be set in the environment or passed as arguments."
            )

        if not datarobot_endpoint.endswith("/"):
            datarobot_endpoint = datarobot_endpoint + "/"

        self._datarobot_endpoint = datarobot_endpoint
        self._datarobot_api_token = datarobot_api_token

        self._oauth_provider_ids = list(oauth_provider_ids)

        if user_agent is None:
            agent_parts = [
                f"DataRobotOAuthProviderClient/{__version__}",
                f"({platform.system()} {platform.release()} {platform.machine()})",
                f"Python-{platform.python_version()}",
            ]

            user_agent = " ".join(agent_parts)

        if http_client is None:
            http_client = httpx.AsyncClient(
                timeout=timeout,
                follow_redirects=follow_redirects,
                verify=ssl_verify,
                headers={
                    "User-Agent": user_agent,
                    "Authorization": f"Bearer {self._datarobot_api_token}",
                },
            )

        self._http_client = http_client

    def register(self, oauth_provider_id: str) -> None:
        self._oauth_provider_ids.append(oauth_provider_id)

    async def __aenter__(self) -> "AsyncOAuth":
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        await self._http_client.aclose()

    async def close(self) -> None:
        """
        Close the HTTP client connection.
        This is an alias for __aexit__ to maintain compatibility with the OAuthComponent interface.
        """
        await self._http_client.aclose()

    async def get_providers(
        self,
    ) -> list[OAuthProvider]:
        """
        Get the list of OAuth providers.

        Raises:
            OAuthServiceNotAuthenticated: If DataRobot credentials are revoked, corrupted or invalid in other way.
            OAuthProviderUnavailable: If the external OAuth provider is currently unavailable.
            OAuthServiceClientErr: For client errors (4xx status codes).
            OAuthServiceUnavailable: For server errors (5xx status codes).
            OAuthServiceError: For any other errors.
        """
        if not self._oauth_provider_ids:
            return []

        resp = await self._http_client.get(
            urljoin(self._datarobot_endpoint, "externalOAuth/providers/"),
            params={
                "ids": self._oauth_provider_ids,
            },
        )

        if not resp.is_success:
            raise_http_exception(resp)

        providers_data = _DROAuthProviders(**resp.json()).data

        return [provider.to_data() for provider in providers_data]

    async def get_authorization_url(
        self,
        *,
        provider_id: str,
        redirect_uri: str,
        state: str | None = None,
        code_verifier: str | None = None,
        **kwargs: Any,
    ) -> OAuthFlowSession:
        """
        Get the authorization URL for the specified OAuth provider.

        Raises:
            OAuthServiceNotAuthenticated: If DataRobot credentials are revoked, corrupted or invalid in other way.
            OAuthProviderUnavailable: If the external OAuth provider is currently unavailable.
            OAuthServiceClientErr: For client errors (4xx status codes).
            OAuthServiceUnavailable: For server errors (5xx status codes).
            OAuthServiceError: For any other errors.
        """
        resp = await self._http_client.post(
            urljoin(
                self._datarobot_endpoint,
                f"externalOAuth/providers/{provider_id}/authorize/",
            ),
            params={
                "redirect_uri": redirect_uri,
                "state": state,
                # we are not supporting other params in the OAuth Providers Service API
            },
        )

        if resp.status_code == 404:
            raise OAuthServiceProviderNotFound(
                f"The OAuth provider was not found: {resp.text}",
                provider_id=provider_id,
            )

        if not resp.is_success:
            raise_http_exception(resp, provider_id=provider_id)

        raw_data = resp.json()
        redirect_data = _DROAuthRedirectData(**raw_data)

        return OAuthFlowSession(
            provider_id=provider_id,
            redirect_uri=redirect_uri,
            state=redirect_data.state,
            authorization_url=redirect_data.redirect_url,
        )

    async def exchange_code(
        self,
        *,
        provider_id: str,
        sess: OAuthFlowSession,
        params: Mapping[str, Any],
        **kwargs: Any,
    ) -> OAuthData:
        """
        Exchange the authorization code for an access token and user profile.

        Raises:
            OAuthServiceNotAuthenticated: If DataRobot credentials are revoked, corrupted or invalid in other way.
            OAuthProviderUnavailable: If the external OAuth provider is currently unavailable.
            OAuthServiceClientErr: For client errors (4xx status codes).
            OAuthServiceUnavailable: For server errors (5xx status codes).
            OAuthServiceError: For any other errors.
        """
        params = dict(params)

        if error := params.get("error"):
            raise OAuthFlowError(
                provider_id=provider_id,
                error_code=error,
                message=params.get("error_description", "OAuth flow failed"),
            )

        resp = await self._http_client.post(
            urljoin(self._datarobot_endpoint, "externalOAuth/providers/callback/"),
            json={
                "providerId": provider_id,
                "code": params.get("code"),
                "state": sess.state,
                # we are not supporting other params like code_verifier in the OAuth Providers Service API
            },
        )

        if resp.status_code == 404:
            raise OAuthServiceProviderNotFound(
                f"The OAuth provider was not found: {resp.text}",
                provider_id=provider_id,
            )

        if not resp.is_success:
            raise_http_exception(resp, provider_id=provider_id)

        raw_data = resp.json()

        oauth_data = _DROAuthData(**raw_data)
        authorization = oauth_data.authorized_provider

        return OAuthData(
            authorization_id=authorization.id,
            provider=authorization.provider.to_data(),
            user_profile=oauth_data.user_info.to_data(),
        )

    async def refresh_access_token(
        self,
        provider_id: str | None = None,
        identity_id: str | None = None,
        scope: str | None = None,
        refresh_token: str | None = None,
    ) -> OAuthToken:
        """
        Refresh the access token for the specified OAuth provider.

        Raises:
            OAuthServiceNotAuthenticated: If DataRobot credentials are revoked, corrupted or invalid in other way.
            OAuthProviderUnavailable: If the external OAuth provider is currently unavailable.
            OAuthServiceClientErr: For client errors (4xx status codes).
            OAuthServiceUnavailable: For server errors (5xx status codes).
            OAuthServiceError: For any other errors.
        """

        if identity_id is None:
            raise OAuthValidationErr(
                "Identity ID is required to obtain new access token via DataRobot API"
            )

        resp = await self._http_client.post(
            urljoin(
                self._datarobot_endpoint,
                f"externalOAuth/authorizedProviders/{identity_id}/token/",
            ),
        )

        if resp.status_code == 404:
            raise OAuthAuthorizationNotFound(
                f"The authorization was not found: {resp.text}",
                provider_id=provider_id,
                authorization_id=identity_id,
            )

        if resp.status_code == 410:
            raise OAuthAuthorizationExpired(
                "The authorization has expired. Please re-authorize the provider.",
                provider_id=provider_id,
                authorization_id=identity_id,
            )

        if not resp.is_success:
            raise_http_exception(resp, provider_id=provider_id, authorization_id=identity_id)

        token_data = _DROAuthTokenData(**resp.json())

        return token_data.to_data()

    async def get_user_info(
        self,
        provider_id: str | None = None,
        identity_id: str | None = None,
        access_token: str | None = None,
    ) -> Profile:
        """
        Get user information from the OAuth provider.

        Raises:
            OAuthServiceNotAuthenticated: If DataRobot credentials are revoked, corrupted or invalid in other way.
            OAuthProviderUnavailable: If the external OAuth provider is currently unavailable.
            OAuthServiceClientErr: For client errors (4xx status codes).
            OAuthServiceUnavailable: For server errors (5xx status codes).
            OAuthServiceError: For any other errors.
        """
        if identity_id is None:
            raise OAuthValidationErr(
                "Identity ID is required to obtain new access token via DataRobot API"
            )

        resp = await self._http_client.get(
            urljoin(
                self._datarobot_endpoint,
                f"externalOAuth/authorizedProviders/{identity_id}/userinfo/",
            ),
        )

        if resp.status_code == 404:
            raise OAuthAuthorizationNotFound(
                f"The authorization was not found: {resp.text}",
                provider_id=provider_id,
                authorization_id=identity_id,
            )

        if resp.status_code == 410:
            raise OAuthAuthorizationExpired(
                "The authorization has expired. Please re-authorize the provider.",
                provider_id=provider_id,
                authorization_id=identity_id,
            )

        if not resp.is_success:
            raise_http_exception(resp, provider_id=provider_id, authorization_id=identity_id)

        user_info = _DROAuthUserProfile(**resp.json())

        return user_info.to_data()


class SyncOAuth(SyncOAuthComponent):
    """
    Synchronous OAuth client for DataRobot OAuth Providers Service API.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the SyncOAuth client.

        Args:
            oauth_provider_ids: A sequence of OAuth provider IDs to register.
            datarobot_endpoint: The DataRobot API endpoint URL. If not provided,
                it will be read from the environment variable DATAROBOT_ENDPOINT.
            datarobot_api_token: The DataRobot API token. If not provided,
                it will be read from the environment variable DATAROBOT_API_TOKEN.
            http_client: An optional HTTP client to use for requests.
                If not provided, a new AsyncClient will be created.
            user_agent: An optional user agent string to use for requests.
            follow_redirects: Whether to follow redirects in HTTP requests.
            timeout: The timeout for HTTP requests.
            ssl_verify: Whether to verify SSL certificates or a custom SSL context.

        Raises:
            ValueError: If the DataRobot endpoint or API token is not provided.
        """
        self._async_client = AsyncOAuth(*args, **kwargs)

    def __enter__(self) -> "SyncOAuth":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        self.close()

    def close(self) -> None:
        """
        Close the HTTP client connection.
        This is an alias for __exit__ to maintain compatibility with the OAuthComponent interface.
        """
        syncify(self._async_client.close)()

    def get_providers(self) -> list[OAuthProvider]:
        return syncify(self._async_client.get_providers)()

    def get_authorization_url(
        self,
        *,
        provider_id: str,
        redirect_uri: str,
        state: str | None = None,
        code_verifier: str | None = None,
        **kwargs: Any,
    ) -> OAuthFlowSession:
        """
        Get the authorization URL for the specified OAuth provider.

        Raises:
            OAuthServiceNotAuthenticated: If DataRobot credentials are revoked, corrupted or invalid in other way.
            OAuthProviderUnavailable: If the external OAuth provider is currently unavailable.
            OAuthServiceClientErr: For client errors (4xx status codes).
            OAuthServiceUnavailable: For server errors (5xx status codes).
            OAuthServiceError: For any other errors.
        """
        return syncify(self._async_client.get_authorization_url)(
            provider_id=provider_id,
            redirect_uri=redirect_uri,
            state=state,
            code_verifier=code_verifier,
            **kwargs,
        )

    def exchange_code(
        self,
        *,
        provider_id: str,
        sess: OAuthFlowSession,
        params: Mapping[str, Any],
        **kwargs: Any,
    ) -> OAuthData:
        """
        Exchange the authorization code for an access token and user profile.

        Raises:
            OAuthServiceNotAuthenticated: If DataRobot credentials are revoked, corrupted or invalid in other way.
            OAuthProviderUnavailable: If the external OAuth provider is currently unavailable.
            OAuthServiceClientErr: For client errors (4xx status codes).
            OAuthServiceUnavailable: For server errors (5xx status codes).
            OAuthServiceError: For any other errors.
        """
        return syncify(self._async_client.exchange_code)(
            provider_id=provider_id,
            sess=sess,
            params=params,
            **kwargs,
        )

    def refresh_access_token(
        self,
        provider_id: str | None = None,
        identity_id: str | None = None,
        scope: str | None = None,
        refresh_token: str | None = None,
    ) -> OAuthToken:
        """
        Refresh the access token for the specified OAuth provider.

        Raises:
            OAuthServiceNotAuthenticated: If DataRobot credentials are revoked, corrupted or invalid in other way.
            OAuthProviderUnavailable: If the external OAuth provider is currently unavailable.
            OAuthServiceClientErr: For client errors (4xx status codes).
            OAuthServiceUnavailable: For server errors (5xx status codes).
            OAuthServiceError: For any other errors.
        """
        return syncify(self._async_client.refresh_access_token)(
            provider_id=provider_id,
            identity_id=identity_id,
            scope=scope,
            refresh_token=refresh_token,
        )

    def get_user_info(
        self,
        provider_id: str | None = None,
        identity_id: str | None = None,
        access_token: str | None = None,
    ) -> Profile:
        """
        Get user information from the OAuth provider.

        Raises:
            OAuthServiceNotAuthenticated: If DataRobot credentials are revoked, corrupted or invalid in other way.
            OAuthProviderUnavailable: If the external OAuth provider is currently unavailable.
            OAuthServiceClientErr: For client errors (4xx status codes).
            OAuthServiceUnavailable: For server errors (5xx status codes).
            OAuthServiceError: For any other errors.
        """
        return syncify(self._async_client.get_user_info)(
            provider_id=provider_id,
            identity_id=identity_id,
            access_token=access_token,
        )
